package ru.dbotthepony.kstarbound.client.gl.shader

import ru.dbotthepony.kstarbound.client.gl.BlendFunc
import ru.dbotthepony.kstarbound.client.gl.GLStateTracker
import ru.dbotthepony.kstarbound.client.gl.GLType
import ru.dbotthepony.kstarbound.client.gl.vertex.GLAttributeList
import ru.dbotthepony.kstarbound.client.gl.vertex.StreamVertexBuilder
import ru.dbotthepony.kstarbound.client.gl.vertex.GeometryType
import ru.dbotthepony.kstarbound.client.gl.vertex.QuadTransformers
import ru.dbotthepony.kvector.arrays.Matrix4f
import ru.dbotthepony.kvector.vector.RGBAColor

private fun GLStateTracker.shaders(name: String): List<GLStateTracker.Shader> {
	return listOf(internalVertex("shaders/$name.vsh"), internalFragment("shaders/$name.fsh"))
}

private fun GLStateTracker.gshaders(name: String): List<GLStateTracker.Shader> {
	return listOf(
		internalVertex(name),
		internalFragment(name),
		internalGeometry(name)
	)
}

class GLLiquidProgram(state: GLStateTracker) : GLShaderProgram(state, state.shaders("liquid"), FORMAT) {
	var baselineColor by F4Uniform("baselineColor")
	var transform by F4x4Uniform("transform")

	val builder by lazy {
		StreamVertexBuilder(state, FORMAT, GeometryType.QUADS, 16384)
	}

	companion object {
		val FORMAT = GLAttributeList.Builder().push(GLType.VEC2F).build()
	}
}

class GLLightProgram(state: GLStateTracker) : GLShaderProgram(state, state.shaders("light"), FORMAT) {
	var baselineColor by F4Uniform("baselineColor")
	var transform by F4x4Uniform("transform")

	val builder by lazy {
		StreamVertexBuilder(state, FORMAT, GeometryType.QUADS, 32)
	}

	companion object {
		val FORMAT = GLAttributeList.Builder().push(GLType.VEC2F).push(GLType.VEC2F).build()
	}
}

class GLColorQuadProgram(state: GLStateTracker) : GLShaderProgram(state, state.shaders("screen_quad"), FORMAT) {
	var color by F4Uniform("color")

	private val builder by lazy {
		val builder = StreamVertexBuilder(state, FORMAT, GeometryType.QUADS, 1)

		builder.builder.begin()
		builder.builder.quad(-1f, -1f, 1f, 1f)
		builder.upload()

		builder
	}

	fun clearAlpha() {
		use()

		color = ALPHA

		val old = state.blend
		val oldFunc = state.blendFunc

		state.blend = true
		state.blendFunc = BlendFunc.ONLY_ALPHA
		builder.draw()
		state.blend = old
		state.blendFunc = oldFunc
	}

	fun clearColor(color: RGBAColor = RGBAColor.WHITE) {
		use()

		this.color = color

		val old = state.blend

		state.blend = false
		builder.draw()
		state.blend = old
	}

	companion object {
		val FORMAT = GLAttributeList.Builder().push(GLType.VEC2F).build()
		val ALPHA = RGBAColor(0f, 0f, 0f, 1f)
	}
}

class GLTextureQuadProgram(state: GLStateTracker) : GLShaderProgram(state, state.shaders("screen_quad_tex"), FORMAT) {
	var texture by IUniform("texture0")

	private val builder by lazy {
		val builder = StreamVertexBuilder(state, FORMAT, GeometryType.QUADS, 1)

		builder.builder.begin()
		builder.builder.quad(-1f, -1f, 1f, 1f, QuadTransformers.uv())
		builder.upload()

		builder
	}

	fun run(texture: Int = 0) {
		use()
		this.texture = texture
		builder.draw()
	}

	companion object {
		val FORMAT = GLAttributeList.Builder().push(GLType.VEC2F).push(GLType.VEC2F).build()
	}
}

class GLTextureBlurredQuadProgram(state: GLStateTracker) : GLShaderProgram(state, state.shaders("screen_quad_tex_blur"), FORMAT) {
	var texture by IUniform("texture0")

	private val builder by lazy {
		val builder = StreamVertexBuilder(state, FORMAT, GeometryType.QUADS, 1)

		builder.builder.begin()
		builder.builder.quad(-1f, -1f, 1f, 1f, QuadTransformers.uv())
		builder.upload()

		builder
	}

	fun run(texture: Int = 0) {
		use()
		this.texture = texture
		builder.draw()
	}

	companion object {
		val FORMAT = GLAttributeList.Builder().push(GLType.VEC2F).push(GLType.VEC2F).build()
	}
}

class GLFlatColorProgram(state: GLStateTracker) : GLShaderProgram(state, state.shaders("flat_color"), FORMAT) {
	var transform by F4x4Uniform("transform")

	val builder by lazy {
		StreamVertexBuilder(state, FORMAT, GeometryType.QUADS, 16384)
	}

	companion object {
		val FORMAT = GLAttributeList.Builder().push(GLType.VEC2F).push(GLType.VEC4F).build()
	}
}

class GLTileProgram(state: GLStateTracker) : GLShaderProgram(state, state.shaders("tile"), FORMAT) {
	var transform by F4x4Uniform("transform")
	var color by F4Uniform("color")
	var texture by IUniform("texture0")

	init {
		transform = Matrix4f.identity()
		color = RGBAColor.WHITE
	}

	companion object {
		val FORMAT = GLAttributeList.Builder().push(GLType.VEC3F, GLType.VEC2F, GLType.FLOAT).build()
	}
}

class GLFontProgram(state: GLStateTracker) : GLShaderProgram(state, state.shaders("font"), GLAttributeList.VERTEX_2D_TEXTURE) {
	var transform by F4x4Uniform("transform")
	var color by F4Uniform("color")
	var texture by IUniform("texture0")

	init {
		transform = Matrix4f.identity()
		color = RGBAColor.WHITE
	}
}

class GLFlatProgram(state: GLStateTracker) : GLShaderProgram(state, state.shaders("flat"), GLAttributeList.VEC2F) {
	var transform by F4x4Uniform("transform")
	var color by F4Uniform("color")

	init {
		color = RGBAColor.WHITE
	}
}

class GLTexturedProgram(state: GLStateTracker) : GLShaderProgram(state, listOf(state.internalVertex("shaders/vertex/texture.glsl"), state.internalFragment("shaders/fragment/texture.glsl")), GLAttributeList.VERTEX_TEXTURE) {
	var transform by F4x4Uniform("transform")
	var texture by IUniform("texture0")

	init {
		transform = Matrix4f.identity()
	}
}

class GLTexturedColoredProgram(state: GLStateTracker) : GLShaderProgram(state, listOf(state.internalVertex("shaders/vertex/texture.glsl"), state.internalFragment("shaders/fragment/texture_color.glsl")), GLAttributeList.VERTEX_TEXTURE) {
	var transform by F4x4Uniform("transform")
	var texture by IUniform("texture0")
	var color by F4Uniform("color")

	init {
		transform = Matrix4f.identity()
		color = RGBAColor.WHITE
	}
}

class GLPrograms(val state: GLStateTracker) {
	val tile by lazy { GLTileProgram(state) }
	val font by lazy { GLFontProgram(state) }
	val flat by lazy { GLFlatProgram(state) }
	val flatColor by lazy { GLFlatColorProgram(state) }
	val liquid by lazy { GLLiquidProgram(state) }
	val light by lazy { GLLightProgram(state) }
	val textured by lazy { GLTexturedProgram(state) }
	val texturedColored by lazy { GLTexturedColoredProgram(state) }

	val viewColorQuad by lazy { GLColorQuadProgram(state) }
	val viewTextureQuad by lazy { GLTextureQuadProgram(state) }
	val viewTextureBlurQuad by lazy { GLTextureBlurredQuadProgram(state) }
}
